/*
  averages.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <list>
#include <functional>
#if __cplusplus < 201103L
#include <boost/function.hpp>
#endif
#include <cstdlib>      // random(), etc
#include <cassert>
#include "averages.hpp"
#include "averages_online.hpp"

template <typename T>
struct identity_function : std::unary_function<T, T> {
  T operator()(const T &x) const
  { return x; }
};

template <typename T>
struct reciprocal_function : std::unary_function<T, T> {
  T operator()(const T &x) const
  { return T(1)/x; }
};

template <typename T>
struct square_function : std::unary_function<T, T> {
  T operator()(const T &x) const
  { return x*x; }
};

//#define USE_LIST_FOR_MOVING_STORAGE
//#define USE_ARBITRARY_GENERALIZED_AVERAGE
//#define USE_ARBITRARY_GENERALIZED_AVERAGE_ONLINE

int main()
{
  const double maximum_permissible_error(1e-14);
  const char *result_names[] = {
    "arithmetic mean",
    "geometric mean",
    "harmonic mean",
    "generalized mean",
    "exponential smoothed average",
  };
  std::vector<double> v(10000);
  generic_average_online<double> *online_averages[] = {
#if !defined(USE_ARBITRARY_GENERALIZED_AVERAGE_ONLINE)
    new arithmetic_average_online<double>(),
    new geometric_average_online<double>(),
    new harmonic_average_online<double>(),
    new generalized_average_online<double>(),
#else
    new arbitrary_generalized_average_online<double>(identity_function<double>(), identity_function<double>()),
    new arbitrary_generalized_average_online<double>(log, exp),
    new arbitrary_generalized_average_online<double>(reciprocal_function<double>(), reciprocal_function<double>()),
    new arbitrary_generalized_average_online<double>(square_function<double>(), sqrt),
#endif
    new exponential_smoothed_average_online<double>(),
  };
  srandomdev();
  for (size_t i=0; i<v.size(); ++i) {
    double x((double)random()/((1L<<31)-1));
    v[i] = x;
    for (size_t j=0; j<sizeof(online_averages)/sizeof(online_averages[0]); ++j)
      online_averages[j]->average(x);
  }
  double results[] = {
#if !defined(USE_ARBITRARY_GENERALIZED_AVERAGE)
    arithmetic_average(v.begin(), v.end()),
    geometric_average(v.begin(), v.end()),
    harmonic_average(v.begin(), v.end()),
    generalized_average(v.begin(), v.end()),
#else
    generalized_average(v.begin(), v.end(), identity_function<double>(), identity_function<double>()),
    generalized_average(v.begin(), v.end(), log, exp),
    generalized_average(v.begin(), v.end(), reciprocal_function<double>(), reciprocal_function<double>()),
    generalized_average(v.begin(), v.end(), square_function<double>(), sqrt),
#endif
    exponential_smoothed_average(v.begin(), v.end()),
  };
  for (size_t j=0; j<sizeof(online_averages)/sizeof(online_averages[0]); ++j) {
    double error(sqrt(pow(online_averages[j]->get_value() - results[j], 2)));
    std::cout << result_names[j] << ":\t" << online_averages[j]->get_value() << "\t" << results[j] << "\t" << error << std::endl;
    assert(error < maximum_permissible_error);
  }

  const char *moving_result_names[] = {
    "moving arithmetic average",
    "moving geometric average",
    "moving harmonic average",
    "moving generalized average",
  };
#if !defined(USE_LIST_FOR_MOVING_STORAGE)
  std::vector<double> m(v.size());
#else
  std::list<double> m;
#endif
  generic_average_online<double> *online_moving_averages[] = {
#if !defined(USE_ARBITRARY_GENERALIZED_AVERAGE_ONLINE)
    new moving_arithmetic_average_online<double>(v.size()),
    new moving_geometric_average_online<double>(v.size()),
    new moving_harmonic_average_online<double>(v.size()),
    new moving_generalized_average_online<double>(v.size()),
#else
    new moving_arbitrary_generalized_average_online<double>(v.size(), identity_function<double>(), identity_function<double>()),
    new moving_arbitrary_generalized_average_online<double>(v.size(), log, exp),
    new moving_arbitrary_generalized_average_online<double>(v.size(), reciprocal_function<double>(), reciprocal_function<double>()),
    new moving_arbitrary_generalized_average_online<double>(v.size(), square_function<double>(), sqrt),
#endif
  };
  for (size_t i=0; i<v.size()*10; ++i) {
    double x((double)random()/((1L<<31)-1));
#if !defined(USE_LIST_FOR_MOVING_STORAGE)
    m[i % v.size()] = x;
#else
    if (!(i < v.size())) m.pop_front();
    m.push_back(x);
#endif
    for (size_t j=0; j<sizeof(online_moving_averages)/sizeof(online_moving_averages[0]); ++j)
      online_moving_averages[j]->average(x);
    if (i < v.size()) continue;
    double moving_results[] = {
#if !defined(USE_ARBITRARY_GENERALIZED_AVERAGE)
      arithmetic_average(m.begin(), m.end()),
      geometric_average(m.begin(), m.end()),
      harmonic_average(m.begin(), m.end()),
      generalized_average(m.begin(), m.end()),
#else
      generalized_average(m.begin(), m.end(), identity_function<double>(), identity_function<double>()),
      generalized_average(m.begin(), m.end(), log, exp),
      generalized_average(m.begin(), m.end(), reciprocal_function<double>(), reciprocal_function<double>()),
      generalized_average(m.begin(), m.end(), square_function<double>(), sqrt),
#endif
    };
    for (size_t j=0; j<sizeof(online_moving_averages)/sizeof(online_moving_averages[0]); ++j) {
      double error(sqrt(pow(online_moving_averages[j]->get_value() - moving_results[j], 2)));
      if (i+1 == v.size()*10)
        std::cout << moving_result_names[j] << ":\t" << online_moving_averages[j]->get_value() << "\t" << moving_results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
  }
  return 0;
}
